# Title     : TODO
# Objective : TODO
# Created by: joshualuthy
# Created on: 8/9/21
# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need)
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}
# check for packages then install if needed.
loadPackages("dplyr","data.table","Matrix","Seurat","sctransform","future","tidyr")
# loadBioc("scran","glmGamPoi")

# library(future)
# check the current active plan
plan()
# change the current plan to access parallelization
plan("multiprocess", workers = 4)
plan()
options(future.seed=TRUE)
# 'Futurized' functions in Seurat
# NormalizeData
# ScaleData
# JackStraw
# FindMarkers
# FindIntegrationAnchors
# FindClusters - if clustering over multiple resolutions
#######################################################################
###### Start the File Import BatchLR method using fread
#######################################################################
# Read CSV from SeqGeq
fastCSVfile <- fread(file= "/Users/joshualuthy/Library/Caches/com.flowjo.flowjovx/SeqGeqRemoteFileCache/zip/3f1767c6b44d065ec1c57873e5ab58a8-18patients_k.geqzip/BatchLR_RunID_CJCA/concat18patients.csv.QCpop.ExtNode.csv",
                     sep = ",", header = TRUE, check.names = FALSE)
featureAnnotationsFrame <- data.frame(fastCSVfile$Gene)
featureAnnos <- as.character(fastCSVfile$Gene)
cellAnnotationsFrame <- data.frame(colnames(fastCSVfile))

if(any(names(fastCSVfile) == "Gene")) {
  cellAnnos <- as.character(colnames(fastCSVfile))
  cellAnnos <- cellAnnos[cellAnnos != "Gene"]

  cellAnnotationsFrame <- data.frame(colnames(fastCSVfile))
  rowsToKeep <- fastCSVfile[, which(names(fastCSVfile) != "Gene")]
  cellAnnotationsFrame <- data.frame(cellAnnotationsFrame[rowsToKeep,])
  colnames(cellAnnotationsFrame) <- paste("CellId")

}
rm(rowsToKeep)

fastCSVfile <- as.data.frame(fastCSVfile)
rownames(fastCSVfile) <- featureAnnos
fastCSVfile <- dplyr::select(fastCSVfile, -Gene)

batchFrame <- fastCSVfile["SampleID",]

# Some data has numbers as cellIDs only, append a string to cellIds, remove on export.
colnames(fastCSVfile) <- paste("AppID", colnames(fastCSVfile), sep="_")
# Transpose the CSV so that the column 'SampleID' can be used for formatting prior to making a Seurat Object
newDF <- data.frame(t(fastCSVfile))
splitDF <- split(newDF, newDF$SampleID)
# now order them from high to low for merging steps
splitDF <- splitDF[order(sapply(splitDF, NROW), decreasing = T)]
# now lets name these by sample name from geqzip( getSamples) or from FJServer and my Noir package...?
# namesToAdd <- getSamples(urlSams) # from Noir
# namesToAdd <- c("Ctrl.csv", "BACT.csv", "Day4-BACT") # This list should come from SeqGeq Java
# alternate to nzchar:is a fast way to find out if elements of a character vector are non-empty strings? ideally would use !nzchar
# names(splitDF)[nzchar(names(splitDF))] <- namesToAdd # ideally would use !nzchar
# Now remove the SampleID column before making Seurat Objects
for(i in seq_along(splitDF)) {
  splitDF[[i]] <- subset(splitDF[[i]], select = -c(SampleID))
}
# Now transpose them to params in cloumns for Seurat
for(i in seq_along(splitDF)){
  splitDF[[i]] <- t(splitDF[[i]])
}
###################################
## Create List of Seurat Objects ##
###################################
# for (i in seq_along(splitDF)) {            # 2. sequence
#   output[[i]] <- CreateSeuratObject(splitDF[[i]]) #, project = namesToAdd[i])      # 3. body
# }
output <<- vector("list", length(splitDF))
for (i in seq_along(splitDF)) {            # 2. sequence
  output[[i]] <- CreateSeuratObject(splitDF[[i]])#, project = paste0("replicate",[[i]]))      # 3. body
}

# After acquiring the data, we first perform standard normalization and variable feature selection.
for (i in seq_along(output)) {
  output[[i]] <- NormalizeData(output[[i]], verbose = FALSE)
}

for (i in seq_along(output)) {
  output[[i]] <- FindVariableFeatures(output[[i]], selection.method = "vst",nfeatures = 2000, verbose = FALSE)
}

######################################
## Create List of Reference Objects ##
######################################
names(output) <- names(splitDF)
rm(splitDF)
# should now clear splitDF from memory to free it (1.5GB for my test dataset)
reference.list <- output[c("15","14")]#,"13","12")]
options(future.globals.maxSize= 891289600) # Had to increase max size of future.globals.maxSize to 850MB (850*1024^2 = 891289600)

# Next, select features for downstream integration, and run PCA on each object in the list,
# which is required for running the rPCA workflow.
myFeatures <- SelectIntegrationFeatures(object.list = reference.list, nfeatures = 2000)
reference.list <- lapply(X = reference.list, FUN = function(x) {
  x <- ScaleData(x, features = myFeatures, verbose = TRUE)
  x <- RunPCA(x, features = myFeatures, verbose = TRUE)
})
# Below is the slow step when not specifying 'reduction' it uses 'CCA'. Lets speed it up with reduction='rpca'
nasal.anchors <- FindIntegrationAnchors(object.list = reference.list, dims = 1:40, reduction = 'rpca')

# Next integrate these into one Seurat Reference Object
# This creates new Assay slot 'integrated' use this downstream workflow
nasal.integrated <- IntegrateData(anchorset = nasal.anchors, dims = 1:40)

nasal.integrated <- ScaleData(nasal.integrated, verbose = FALSE)
# Now run PCA on the integrated dataset for use in clustering and umap
nasal.integrated <- RunPCA(nasal.integrated, assay = 'integrated')


nasal.integrated <- FindNeighbors(nasal.integrated, dims = 1:15)
nasal.integrated <- FindClusters(nasal.integrated, resolution = 0.2)#, graph.name = "integrated_nn")
#TODO run UMAP or SPCA?
nasal.integrated <- RunUMAP(nasal.integrated,
                            assay = 'integrated',
                            reduction = 'pca',
                            reduction.name = 'umapIntegrated',
                            reduction.key = 'intUMAP_',
                            return.model = TRUE,
                            nn.name = 'pca.annoy.neighbors',
                            dims = 1:30,
                            verbose = FALSE)


# nasal.integrated <- RunSPCA(nasal.integrated, assay = 'integrated', graph = 'integrated_snn')


# compute a cached neighbor index
nasal.integrated <- FindNeighbors(
  object = nasal.integrated,
  reduction = "pca",
  dims = 1:50,
  graph.name = "pca.annoy.neighbors",
  k.param = 50,
  cache.index = TRUE,
  return.neighbor = TRUE,
  l2.norm = TRUE
)

###################################
## Create List of Query Objects ##
###################################
query.list <- output[c("0","1")]#,"2","3","4","5","6","7","8","9","10","11")]

for (i in seq_along(query.list)) {
  Idents(query.list[[i]]) <- paste0("replicate", i )
}

# anchors2 <- list()
anchors2 <- vector("list", length(query.list))
for (i in seq_along(query.list)) {
  anchors2[[i]] <- FindTransferAnchors(reference = nasal.integrated,
                                       query = query.list[[i]],
                                       k.filter = NA,
                                       reference.reduction = "pca",
                                       reference.neighbors = "pca.annoy.neighbors",
                                       dims = 1:50)
}

# Next, map each dataset
for (i in 1:length(query.list)) {
  query.list[[i]] <- MapQuery(
    anchorset = anchors2[[i]],
    query = query.list[[i]],
    reference = nasal.integrated,
    refdata = list(
      clusters = "seurat_clusters",
      predictedRNA = "RNA"),
    reference.reduction = "pca",
    reduction.model = "umapIntegrated"
  )
}


# p1 <- DimPlot(query.list[[1]], group.by = "predicted.clusters",reduction = 'ref.umap', label.size = 3)
# p2 <- DimPlot(query.list[[2]], group.by = "predicted.clusters",reduction = 'ref.umap', label.size = 3)
#
#
# refPlot <- DimPlot(nasal.integrated, reduction = "umapIntegrated")
# refPlot + p1 + p2 + plot_layout(guides = "auto")
#
# integratedPlot <- DimPlot(nasal.integrated, reduction = "pca")
# integratedPlot

# Now we can merge results
# Merge the batches
# TODO here i had to increase memory on my win10 16GB RAM to '20000' with memory.limit()
if(.Platform$OS.type == "unix") {
} else {
  memory.limit()
  memory.limit(20000)
}

# library(benchmarkme)
#
# ## Return the machine CPU
# cat("Machine:     "); print(get_cpu()$model_name)
#
# ## Return number of true cores
# cat("Num cores:   "); print(detectCores(logical = FALSE))
#
# ## Return number of threads
# cat("Num threads: "); print(detectCores(logical = TRUE))
#
# ## Return the machine RAM
# cat("RAM:         "); print (get_ram()); cat("\n")
nasal.merged <- merge(query.list[[1]], query.list[2:length(query.list)], merge.dr = "ref.umap",)
mergedPlot <- DimPlot(nasal.merged, reduction = "ref.umap",
                      # group.by = "predicted.clusters",
                      label = TRUE, repel = TRUE, label.size = 5)
mergedPlot# + refPlot + p1 + p2 + plot_layout(guides = "collect")

FeaturePlot(query.list[[2]], features = c("APOD", "CD52","COL1A2","SFRP2", "SH3BGRL3"),  reduction = "ref.umap", cols = c("lightgrey", "darkred"), ncol = 3) & theme(plot.title = element_text(size = 10))

# SplitDotPlotGG has been replaced with the `split.by` parameter for DotPlot
myfeatures <- c("APOD", "CD52","COL1A2","SFRP2", "SH3BGRL3")
DotPlot(nasal.integrated, features = myfeatures, split.by = "orig.ident") + RotatedAxis()
# Violin plots can also be split on some variable. Simply add the splitting variable to object
# metadata and pass it to the split.by argument
VlnPlot(nasal.integrated, features = myfeatures, pt.size = 0)
# we want ref.umap and predicted.clusters from the merged object
merged.umap <- as.data.frame(nasal.merged@reductions[["ref.umap"]]@cell.embeddings)
merged.umap[,3] <- as.data.frame(gsub("AppID_*", "", rownames(merged.umap)))
colnames(merged.umap)[3] <- "CellId"
merged.umap <- merged.umap %>% dplyr::select(CellId, everything())
merged.clusters <- as.data.frame(nasal.merged@meta.data[["predicted.clusters"]])
colnames(merged.clusters) <- "predicted.clusters"
mergedResults <- cbind(merged.umap, merged.clusters)
# export the results
fwrite(mergedResults, file = "/Users/joshualuthy/mergedResults.csv", sep = ",", row.names = FALSE)
# write.table(merged.umap.clusters, file = "/Users/joshualuthy/merged.umap.clusters.csv", sep = ',', append = T, quote = F, row.names = T, col.names = T , na="")
