
# Script modified from Satija lab's work for the R Seurat package.
# Script modified by Ian Taylor of BD Life Science - Informatics
# For use w/ SeqGeq 
# Seurat is developed and maintained by the Satija lab. 
# Code is  released in GPL (3.0) license. 

#Disable error messaging in v1.0, prevent early exit.
options(warn = - 1)


## INSTALL REQUIRED PACKAGES ##
#######################

## Note, this old way doesn't work:
# if (!requireNamespace("BiocManager", quietly = TRUE)) install.packages("BiocManager")
# BiocManager::install(c("flowCore", "FlowSOM"))

## TRY TO AUTO INSTALL ##

# Install function for packages
packages<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="https://cloud.r-project.org")
  }else{
    require(x,character.only=TRUE)
  }
}

fCorePack<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    BiocManager::install(x)
    require(x,character.only=TRUE)
  }
}

# Get required packages #
packages(gplots)
packages(RColorBrewer)
packages(dplyr)
packages(Matrix)
packages(Seurat)
packages(cowplot)
packages(reshape2)
packages(ggplot2)
packages(tidyr)
packages(sctransform)
packages(tools)
packages(xfun)
fCorePack(multtest)


## Other Functions ##

fig0 <- function(DF){
VlnPlot(DF, features = c("nFeature_RNA", "nCount_RNA", "percent.mt"), ncol = 3)
ggsave(file=paste0(outfileDirectory, "/qcPlot.png"))
}

fig1 <- function(DF, feat){
  FeaturePlot(DF, features = feat)
  ggsave(file=paste0(outfileDirectory, "/cellTypes.png"))
}

fig2 <- function(DF) {
  ggplot(DF, aes(y = Features, x = Type, fill = Fill)) +
    theme(axis.text.x = element_text(angle = 45, hjust = 1, vjust = 0.9, size = 22),
          axis.text.y = element_text(hjust = 0.9, vjust = 0.5, size = 22),
          legend.title = element_text(size = 18, face="bold"),
          legend.key.size = unit(4,"line"),
          legend.text = element_text(size = 18)) +
    geom_tile() +
     theme(plot.margin = margin(1, 1, 1, 3, "cm")) +
    labs(x = "", y = "") +
    scale_fill_viridis_c(
      name = "Rank Cell Call")
  ggsave(file=paste0(outfileDirectory, "/CellType_Heatmap.png"), height=nrow(tMatrix)*1, width=ncol(tMatrix)*3, bg="white")
}

is.defined <- function(sym) {
  sym <- deparse(substitute(sym))
  env <- parent.frame()
  exists(sym, env)
}

loadRData <- function(fileName){
  #loads an RData file, and returns it
  load(fileName)
  get(ls()[ls() != "fileName"])
}

######################
######################
getSeuratVersion <- function(){
  pkgVersion <- getNamespaceVersion("Seurat")
  pkgVersion <- file_path_sans_ext(pkgVersion)
  pkgVersion <- as.numeric(pkgVersion)
}
yourVersion <- getSeuratVersion()

#%%%%%%%%%%%%%%%%%%%%%%%#
## Read Data from Java ##
#%%%%%%%%%%%%%%%%%%%%%%%#

# runID <- "SG_RUNID"
runID <- ""
model <- "SG_MODEL"

# Read CSV from SeqGeq
inputString <- "SG_ROBJ"
extenShaun <- file_ext(inputString)
outfileDirectory = "SG_OUTPUT_DIRECTORY_PATH"

# CLR_NORMALIZE = CLRNORM

# Read CSV from SeqGeq
if (extenShaun == "rda") {
  import = loadRData(inputString)
  # seuratSeqGeq <- importRobjRDA
} else if (extenShaun == "rds") {
  import <- readRDS(inputString)
} else if (extenShaun == "Robj"){
  import <- get(load(inputString, verbose=TRUE))
} else {
  writeLines("Could not load the data file specified. Make sure output from Seurat is in RDA or RDS format.")
}

## Deal with data like the Sade Feldman example:
if (class(import) == "SingleCellExperiment") {

  writeLines("You've got SCE dataset")

  ###########################################
  ######## What is the raw Matrix?: #########
  if ( !is.null(import@assays$data) ) {
    dataQ <- import@assays$data
    dataMatrix <- dataQ@listData[["counts"]]
    DM <- as.matrix(dataMatrix)
    write.csv(DM, paste0(outfileDirectory,"/SCE_DataMatrix.csv"))
  } else {
    writeLines("Could not load this SingleCellExperiment. Contact: seqgeq@bd.com")
  }


  ###########################################
  ######## What are the DPs?: #########
  DP <- c()
  ##TODO: add options for wnnPCs and wnnUMAP params below
  if ( !is.null(import@reducedDims) ) {
    dimRedux <- import@reducedDims

    namer <- names(dimRedux@listData)
    namerIsPC <- grep("PCA", namer, value = TRUE)
    ###PCA###
    if (!isEmpty(namerIsPC)) {
      PC.DR <- dimRedux@listData$PCA
      pc.dr <- as.matrix(PC.DR)
      DP <- cbind(DP, pc.dr)
    }

    namerIsUMAP <- grep("UMAP", namer, value = TRUE)
    ###UMAP###
    if ( !isEmpty(namerIsUMAP) ) {
      U.DR <- dimRedux@listData$UMAP
      u.dr <- as.matrix(U.DR)
      DP <- cbind(DP, u.dr)
    }

    namerIsTSNE <- grep("TSNE", namer, value = TRUE)
    ###UMAP###
    if ( !isEmpty(namerIsTSNE) ) {
      TSNE.DR <- dimRedux@listData$TSNE
      colnames(TSNE.DR) <- c("tSNE_X", "tSNE_Y")
      tsne.dr <- as.matrix(TSNE.DR)
      DP <- cbind(DP, tsne.dr)
    }
  }

  ###CLUSTERS###
  if ( !isEmpty(import@colData@listData[["SampleID"]]) ) {
    install.packages("sprof", repos="http://R-Forge.R-project.org")
    library(sprof)
    Clusters <- import@colData@listData[1:10]
    clust <- list.as.matrix(Clusters)
    DP <- cbind(DP, clust)
  }

  if (!is.null(DP)) {
    rownames(DP) <- rownames(t(dataMatrix))
    newMat <- as.matrix(rownames(t(dataMatrix)))
    colnames(newMat) <- "CellID"
    DP <- cbind(newMat, DP)
    write.csv(DP, paste0(outfileDirectory,"/DerivedPar.csv"), quote = F, row.names = F)
  }




    ## Deal with the Seurat from SeqGeq example:
} else if (class(import) == "Seurat"){

    writeLines("You've got a Seurat dataset")

    ## TODO - What's the cellID going to be?

    ###########################################
    ######## What is the raw Matrix?: #########

    # We need to address the Dimnames when ADT are normalized
    # Also, deal with down-regulated parameters

    isNorm <- F
    # if (CLR_NORMALIZE){
    # import <- Seurat::NormalizeData(obj.integrated, assay = "ADT", normalization.method = "CLR")
    # }

    hasAb <- F
    if (!is.null(import@assays$ADT)) {

        if(!is.null(import@assays$ADT@data)) {
            rownames(import@assays$ADT@data) <- lapply (rownames(import@assays$ADT@data), paste0, " (Ab)")
            isNorm <- T
        } else if(!is.null(import@assays$ADT@data@Dimnames[1])){
            import@assays$ADT@data@Dimnames[1] <- lapply (import@assays$ADT@data@Dimnames[1], paste0, " (Ab)")
        }
        AB <- import@assays$ADT@data
        hasAb <- T
    }

    if (!is.null(import@assays$integrated)){
        dataMatrix <- import@assays$integrated@data
        DM <- as.data.frame(t(dataMatrix))

        if(hasAb){
            DM <- cbind(DM,t(AB))
        }

        # if(!is.null(import@meta.data$orig.ident)){
        #     sampleNames <- import@meta.data$orig.ident
        #     row.names(DM) <- paste0(sampleNames,"-", row.names(DM))
        # }
        # 
        # if(!is.null(import@meta.data$SingleR_Labels)){
        #     cellTypes <- import@meta.data$SingleR_Labels
        #     row.names(DM) <- paste0(row.names(DM),"-",cellTypes)
        # }

        write.csv(DM, paste0(outfileDirectory,"/CCA_Seurat_DataMatrix.csv"))

    } else if ( !is.null(import@assays[["RNA"]]) ) {
        dataMatrix <- import@assays[["RNA"]]@data
        DM <- as.matrix(dataMatrix)

        if(hasAb){
            newDM <- cbind(t(DM),t(AB))
        } else {
            newDM <- t(DM)
        }
        row.names(newDM) <- gsub("string_","", row.names(newDM))

        if(!is.null(import@meta.data$Population) && !is.null(import@meta.data$orig.ident)){
            prefix <- import@meta.data$orig.ident
            sampleNames <- gsub("-","neg",import@meta.data$Population)
            row.names(newDM) <- paste0(prefix,"_",sampleNames,"-", row.names(newDM))
            # }else
            # if(!is.null(import@meta.data$orig.ident)){
            #   sampleNames <- import@meta.data$orig.ident
            #   row.names(DP) <- paste0(sampleNames,"-", row.names(DP))
        } else if(!is.null(import@meta.data$Population)){
            sampleNames <- gsub("-","neg",import@meta.data$Population)
            row.names(newDM) <- paste0(sampleNames,"-", row.names(newDM))
            # }else
            # if(!is.null(import@meta.data$orig.ident)){
            #   sampleNames <- import@meta.data$orig.ident
            #   row.names(DP) <- paste0(sampleNames,"-", row.names(DP))
        } else if(!is.null(import@meta.data$orig.ident)){
            sampleNames <- import@meta.data$orig.ident
            row.names(newDM) <- paste0(sampleNames,"-", row.names(newDM))
        }

        if(!is.null(import@meta.data$SingleR_Labels_Fine)){
            cellTypes <- import@meta.data$SingleR_Labels_Fine
            row.names(newDM) <- paste0(row.names(newDM),"-",cellTypes)
        }else if(!is.null(import@meta.data$SingleR_Labels)){
            cellTypes <- import@meta.data$SingleR_Labels
            row.names(newDM) <- paste0(row.names(newDM),"-",cellTypes)
        }

        DMRows <- as.data.frame(rownames(newDM))
        colnames(DMRows) <- "CellId"
        
        newDM <- cbind(DMRows,newDM)
        
        write.csv(newDM, file = paste0(outfileDirectory,"/Seurat_DataMatrix.csv"), na = "0", quote = F, row.names = F )
    }

    ###########################################
    ######## What are the DPs?: #########

    DP <- c()

    if ( !is.null(import@reductions) ) {
        dimRedux <- import@reductions

        if (!is.null(import@meta.data$orig.ident)){
            SampleId <- as.factor( import@meta.data$orig.ident )
            # SampleId <- import@meta.data$orig.ident
            DP <- cbind(DP, SampleId)
        }

        ###PCA###
        if ( exists("pca", where=dimRedux) ) {
            PC.DR <- dimRedux[["pca"]]@cell.embeddings
            pc.dr <- as.matrix(PC.DR)
            DP <- cbind(DP, pc.dr)
        }

        ###UMAP###
        if ( exists("umap", where=dimRedux) ) {
            U.DR <- dimRedux[["umap"]]@cell.embeddings
            u.dr <- as.matrix(U.DR)
            DP <- cbind(DP, u.dr)
        }

        ###TSNE###
        if ( exists("tsne", where=dimRedux) ) {
            TSNE.DR <- dimRedux[["tsne"]]@cell.embeddings
            tsne.dr <- as.matrix(TSNE.DR)
            DP <- cbind(DP, tsne.dr)
        }
    }

    ###CLUSTERS###

    if ( !is.null(import@active.ident) ) {
        CLUST <- import@active.ident
        clust <- as.matrix(CLUST)
        colnames(clust) <- "Seurat_Cluster"
        DP <- cbind(DP, clust)
    }

    if (!is.null(DP)) {

        row.names(DP) <- gsub("string_","", row.names(DP))

        if(!is.null(import@meta.data$Population) && !is.null(import@meta.data$orig.ident)){
            prefix <- import@meta.data$orig.ident
            sampleNames <- gsub("-","neg",import@meta.data$Population)
            row.names(DP) <- paste0(prefix,"_",sampleNames,"-", row.names(DP))

        }else if(!is.null(import@meta.data$Population)){
            sampleNames <- gsub("-","neg",import@meta.data$Population)
            row.names(DP) <- paste0(sampleNames,"-", row.names(DP))
        }else if(!is.null(import@meta.data$orig.ident)){
            sampleNames <- import@meta.data$orig.ident
            row.names(DP) <- paste0(sampleNames,"-", row.names(DP))
        }

        if(!is.null(import@meta.data$SingleR_Labels_Fine)){
            cellTypes <- import@meta.data$SingleR_Labels_Fine
            row.names(DP) <- paste0(row.names(DP),"-",cellTypes)
        }else if(!is.null(import@meta.data$SingleR_Labels)){
            cellTypes <- import@meta.data$SingleR_Labels
            row.names(DP) <- paste0(row.names(DP),"-",cellTypes)
        }

        DPRows <- as.data.frame(rownames(DP))
        colnames(DPRows) <- "CellId"
        
        DP <- cbind(DPRows,DP)
        
        write.csv(DP, file = paste0(outfileDirectory,"/DerivedParameters.csv"), na = "0", quote = F, row.names = F )
    }

    #%%%%%%%%%%%%%%%%%%%%%%%%%%%#
    ##     Total Meta Data     ##
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%#
    
    if(!is.null(import@meta.data )){
      
      df <- import@meta.data
      
      # df[is.na(as.data.frame(df))] <- 0
      # df$CellBarcode <- row.names(df)
      
      # row.names(df) <- gsub("string_","", df$CellBarcode)
      
      if(!is.null(import@meta.data$Population) && !is.null(import@meta.data$orig.ident)){
        prefix <- import@meta.data$orig.ident
        sampleNames <- gsub("-","neg",import@meta.data$Population)
        row.names(df) <- paste0(prefix,"_",sampleNames,"-", row.names(df))
        
      }else if(!is.null(import@meta.data$Population)){
        sampleNames <- gsub("-","neg",import@meta.data$Population)
        row.names(df) <- paste0(sampleNames,"-", row.names(df))
      }else if(!is.null(import@meta.data$orig.ident)){
        sampleNames <- import@meta.data$orig.ident
        row.names(df) <- paste0(sampleNames,"-", row.names(df))
      }
      
      if(!is.null(import@meta.data$SingleR_Labels_Fine)){
        cellTypes <- import@meta.data$SingleR_Labels_Fine
        row.names(df) <- paste0(row.names(df),"-",cellTypes)
      }else if(!is.null(import@meta.data$SingleR_Labels)){
        cellTypes <- import@meta.data$SingleR_Labels
        row.names(df) <- paste0(row.names(df),"-",cellTypes)
      }
      
      df <- (gsub(","," : ",as.matrix(df)))
      dfRows <- as.data.frame(rownames(df))
      colnames(dfRows) <- "CellId"
      
      df <- cbind(dfRows,df)
      
      write.csv(df, file = paste0(outfileDirectory,"/Total_MetaData.csv"), na = "0", quote = F, row.names = F )
      
    }
    
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%#
    ## OPTIONS A USER CAN FLAG ##
    #%%%%%%%%%%%%%%%%%%%%%%%%%%%#

  ## NA

  #%%%%%%%%%%%%%%%%%%%%%%%#
  #%%%%%%%%%%%%%%%%%%%%%%%#

  # Log normalize data, add scaling factor

  ## NA

  #%%%%%%%%%%%%%%%%%%%%%%%#
  #%%%%%%%%%%%%%%%%%%%%%%%#



  #%%%%%%%%%%%%%%%%%%%%%%%#
  ## Run DimRedux ##
  #%%%%%%%%%%%%%%%%%%%%%%%#

  ## NA


  ## Find Clusters ##

  ## NA

  #%%%%%%%%%%%%%%%%%%%%%%%#
  #%%%%%%%%%%%%%%%%%%%%%%%#



  #%%%%%%%%%%%%%%%%%%%%%%%#
  ## Clustering ##
  #%%%%%%%%%%%%%%%%%%%%%%%#

  ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
  ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
  # Write out DEG sets #
  ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##

  # findClusterMethod = "negbinom"

  import <- RenameCells(import, new.names = paste0("string_",Cells(import)))
  ## Run test to find top markers for each cluster found in KNN
  import.markers <- FindAllMarkers(object = import, only.pos = F, min.pct = 0.25, test.use = "wilcox", verbose = TRUE, random.seed = FALSE)
  # import.markers <- FindAllMarkers(object = import, only.pos = TRUE, min.pct = 0.25, test.use = findClusterMethod, verbose = TRUE, random.seed = FALSE)
  num.clust <-  length(levels(import@meta.data[["seurat_clusters"]]))

  ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##

  ## UP ##

  for (n in 1:num.clust){

    if(yourVersion < 4.0 ){
      upDEGr <- subset(import.markers, (import.markers$cluster == (n)))
      upDEGr <- subset(upDEGr, upDEGr$avg_logFC > 0)
      upFC <- upDEGr$avg_logFC
    }else{
      upDEGr <- subset(import.markers, (import.markers$cluster == (n)))
      upDEGr <- subset(upDEGr, upDEGr$avg_log2FC > 0)
      upFC <- upDEGr$avg_log2FC
    }

    qVal <- upDEGr$p_val_adj
    feat <- upDEGr$gene

    upDEG <- cbind(feat, upFC, qVal)
    # downFCnQVal <- subset(DEGr, qValTH$Stats....foldChange.. < fcTinv)

    nam <- paste("Up_in_Cluster", n, sep = "_")
    # namAlt <- paste("Down_in", catPar, n, sep = "_")

    assign(nam, upDEG)
    # assign(namAlt, downFCnQVal)

    namGS <- paste0("/Up_in_Cluster", "_", n, "_", runID, ".csv")
    fGSnam <- paste0(outfileDirectory,namGS)
    write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
    write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(paste0("$GeneSetName,Up in Cluster ", n, " RunID: ", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(paste0("$GeneSetDescription,Generated in the Seurat analysis pipeline. Fold-Change > ",min(upFC),"; q-Val (FDR) &lt; ", max(qVal)), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table("[Genes],Log2 Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(upDEG, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")
  }


  ## DOWN ##
  ## decommission this until its better understood how to do this in Seurat:

  # truthy <- T
  # if (!truthy) {

  for (n in 1:num.clust){

    if(yourVersion < 4.0 ){
      downDEGr <- subset(import.markers, (import.markers$cluster == (n)))
      downDEGr <- subset(downDEGr, downDEGr$avg_logFC < 0)
      downFC <- downDEGr$avg_logFC
    }else{
      downDEGr <- subset(import.markers, (import.markers$cluster == (n)))
      downDEGr <- subset(downDEGr, downDEGr$avg_log2FC < 0)
      downFC <- downDEGr$avg_log2FC
    }

    qVal <- downDEGr$p_val_adj
    feat <- downDEGr$gene

    downDEG <- cbind(feat, downFC, qVal)

    nam <- paste("Down_in_Cluster", n, sep = "_")
    # namAlt <- paste("Down_in", catPar, n, sep = "_")

    assign(nam, downDEG)
    # assign(namAlt, downFCnQVal)

    namGS <- paste0("/Down_in_Cluster", "_", n, "_", runID, ".csv")
    fGSnam <- paste0(outfileDirectory,namGS)
    write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
    write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(paste0("$GeneSetName,Down in Cluster ", n, " RunID: ", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(paste0("$GeneSetDescription,Generated in the Seurat analysis pipeline. Fold-Change &lt; ",min(downFC),"; q-Val (FDR) &lt; ", max(qVal)), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table("[Genes],Log2 Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
    write.table(downDEG, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")
  }

  # }

  write.csv(import.markers, file = paste0(outfileDirectory, "Wilcox", "_Summary.csv"), quote = FALSE, row.names = FALSE)
}

##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##

write.table(num.clust, paste0( outfileDirectory, "/Estimated_k" ), row.names = F, col.names = F)

#%%%%%%%%%%%%%%%%%%%%%%%#
## Classification ##
#%%%%%%%%%%%%%%%%%%%%%%%#

## NOTE: This set of parameters is very tricky to find - pMonocytes could be better identified by BCL2A1.
#        The logic for this is very basic - and it requires that a feature be absolutely uniqe to the population of interest.
#        Would benefit from some more advanced boolean logic down the line... (decision trees)
#        Kind of a cool idea none the less.

# TODO - Allow researchers to import their own classification parameters/annotations

if(model != "None"){
  if (!is.null(model)){
    classifyClusters <- T
  } else {
    classifyClusters <- F
  }

  if (classifyClusters) {

    if(model == "huPBMC") {
      ClassGeneSet = c("IL7R", "LYZ", "MS4A1", "CD8B", "CXCL16", "KLRF1", "FCER1A", "PPBP", "GZMK")
    } else if (model == "mmLiver") {
      ClassGeneSet = c("Cd19","Gzma","Cd8a","Cd38","Spp1","F13a1","Cxcr6","Clec4d","Flt3") # ,"","")
    }

    if (length(import@assays[["RNA"]]@scale.data) == 0){

      if (!is.null(import@assays[["SCT"]]) ) {
        ClassGeneSet_rd = Reduce(intersect, list(ClassGeneSet, row.names(import@assays[["SCT"]]@scale.data)))
      } else {
        ClassGeneSet_rd = Reduce(intersect, list(ClassGeneSet, row.names(import@assays$integrated@scale.data)))
      }

    } else {
      ClassGeneSet_rd = Reduce(intersect, list(ClassGeneSet, row.names(import@assays[["RNA"]]@scale.data)))
    }

    geneSize = length(ClassGeneSet)

    foo <- vector(mode="list", length=(geneSize))

    if(model == "huPBMC") {
      names(foo) <- c(
      "IL7R",
      "LYZ",
      "MS4A1",
      "CD8B",
      "CXCL16",
      "KLRF1",
      "FCER1A",
      "PPBP",
      "GZMK")
      foo[[1]] <- "IL7R+ T4 cells";
      foo[[2]] <- "LYZ+ cMonocytes";
      foo[[3]] <- "MS4A1+ B cells";
      foo[[4]] <- "CD8b+ T8n cells";
      foo[[5]] <- "CXCL16+ pMonocytes";
      foo[[6]] <- "KLRF1+ NK cells";
      foo[[7]] <- "FCER1A+ Dendritic cells"
      foo[[8]] <- "PPBP+ Megakaryocytes"
      foo[[9]] <- "GZMK+ T8m cells"
    } else if (model == "mmLiver") {
      names(foo) <- c(
      "Cd19",
      "Gzma",
      "Cd8a",
      "Cd38",
      "Spp1",
      "F13a1",
      "Cxcr6",
      "Clec4d",
      "Flt3")
      # ,
      # "")
      foo[[1]] <- "Cd19+ B cells";
      foo[[2]] <- "Gzma+ NK cells";
      foo[[3]] <- "Cd8a+ T8 cells";
      foo[[4]] <- "Cd38+ Hepatic Stellate cells";
      foo[[5]] <- "Spp1+ LSEndothelial cells";
      foo[[6]] <- "F13a1+ Hepatocytes";
      foo[[7]] <- "Cxcr6+ T4 cells"
      foo[[8]] <- "Clec4d+ Kupffer cells"
      foo[[9]] <- "Flt3+ Hepatic Stellate cells"
      # foo[[10]] <- "+  cells"
    } else if (model == "mmLN-Spleen") {
      names(foo) <- c(
      "Gzma",
      "Ms4a1",
      "Cd8a",
      "Slfn1",
      "Fcgr3",
      "Cxcl16",
      "Vcam1",
      "Retnlg")
      foo[[1]] <- "Gzma+ NK cells";
      foo[[2]] <- "Ms4a1+ B cells";
      foo[[3]] <- "Cd8a+ T8 cells";
      foo[[4]] <- "Slfn1+ T4 cells";
      foo[[5]] <- "Fcgr3 cMonocytes";
      foo[[6]] <- "Cxcl16+ pMonocytes";
      foo[[7]] <- "Vcam+ Endothelial cells"
      foo[[8]] <- "Retnlg+ Macrophage cells"
      # foo[[9]] <- ""
    }

    if (length(import@assays[["RNA"]]@scale.data) == 0){
      if (!is.null(import@assays[["SCT"]]) ) {
        subsetDF = import@assays[["SCT"]]@scale.data
      } else {
        subsetDF = import@assays$integrated@scale.data
      }
    } else {
      subsetDF = import@assays[["RNA"]]@scale.data
    }
    subsetList = import@active.ident
    subDF = rbind(subsetDF, subsetList)
    subDFT = as.data.frame(t(subDF))

    clusterTypesMatrix <- data.frame(matrix(ncol = length(ClassGeneSet_rd), nrow = max(as.list.data.frame(subsetList))))
    colnames(clusterTypesMatrix) <- ClassGeneSet_rd
    rownames(clusterTypesMatrix) <- paste(rownames(clusterTypesMatrix),"Cluster",sep="")
    orderedDF = subDFT[order(subDFT$subsetList),]
    for(gene in colnames(clusterTypesMatrix))
    {
      # Test
      # gene <- "IL7R"

      print(gene)
      lowerBound = -1
      upperBound = -1
      onLevel = 999
      currentMedian = -10

      for(row in 1:length(orderedDF$subsetList))
      {
        # row <- 442

        currentLevel = orderedDF[row,"subsetList"]
        if(onLevel== 999)
        {
          print(currentLevel)
          onLevel = currentLevel
          lowerBound = 0
          upperBound = row
        }
        else if(currentLevel > onLevel){
          upperBound = row - 1
          differ = upperBound - lowerBound
          currentMedian = sum(orderedDF[lowerBound:upperBound,gene])/differ
          lowerBound = upperBound
          index = currentLevel -1
          clusterTypesMatrix[index, gene] = currentMedian
          onLevel = currentLevel
          print(currentLevel)
        }
      }
      upperBound = row
      currentMedian = median(orderedDF[lowerBound:upperBound,gene])
      index = currentLevel
      clusterTypesMatrix[index, gene] = currentMedian
      onLevel = currentLevel
      print(currentLevel)
    }

    clusterTypesMatrixDF = clusterTypesMatrix
    cellCalls = data.frame(matrix(nrow=length(rownames(clusterTypesMatrixDF)), ncol=1))
    colnames(cellCalls) = "CellCall"
    count = 1
    for(cluster in rownames(clusterTypesMatrixDF))
    {

      #print(length(colnames(clusterTypesMatrixDF)))
      if(length(colnames(clusterTypesMatrixDF)) < 1)
      {
        break
      }

      meen <- rowMeans(as.matrix(clusterTypesMatrixDF[cluster,]))
      std <- abs(sd(as.matrix(clusterTypesMatrixDF[cluster,])))

      tst <- meen + 1.5*std

      if (max(clusterTypesMatrixDF[cluster,]) > tst) {

        cellCalls[count,1] = foo[[names(which.max(clusterTypesMatrixDF[cluster,]))]]
        #print(which.max(clusterTypesMatrixDF[cluster,]))
        print(foo[[(which.max(clusterTypesMatrixDF[cluster,]))]])
        #clusterTypesMatrixDF = clusterTypesMatrixDF[,-c(which.max(clusterTypesMatrixDF[cluster,]))]
      } else {
        cellCalls[count,1] = "Unknown"
        print(cellCalls[count,1])
      }
      count = count + 1
    }
    currentCount=1
    for(cell in cellCalls$CellCall)
    {
      rownames(clusterTypesMatrix)[[currentCount]] = paste0("Cluster ", currentCount, " ", cell)
      currentCount = currentCount +1
    }

    lenDF = length(foo) -1
    df <- data.frame(matrix(unlist(foo), nrow=length(foo), byrow=T),stringsAsFactors=FALSE)
    dff <- data.frame(matrix(unlist(names(foo)), nrow=length(foo), byrow=T),stringsAsFactors=FALSE)
    dfff <- cbind(df, dff)
    colnames(dfff) <- c("Type","Gene")

    write.csv(rownames(clusterTypesMatrix), file = paste0(outfileDirectory, "/CellType_Dict.csv"), quote = FALSE, col.names = F, row.names = F)

    XtraCol <- as.data.frame(rownames(clusterTypesMatrix))
    colnames(XtraCol) <- c("Type")
    tMatrix <- cbind(XtraCol, clusterTypesMatrix)

    ## Create Figures ##
    if(ncol (tMatrix) > 2){
      hmDF <- gather(tMatrix, "Features", "Fill", 2:ncol(tMatrix)) # :nrow(clusterTypesMatrix))
      fig2(hmDF)
    }
    fig0(import)
    fig1(import, ClassGeneSet_rd)
    print(" Saved Images ")

  } else {

    ## WE DON'T WANT CLASSIFICATION ##
    # ... but we still need a file to name clusters for SeqGeq... #
    subsetList = import@active.ident
    clusterTypesMatrix <- data.frame(matrix(ncol = 1, nrow = max(as.list.data.frame(subsetList))))
    colnames(clusterTypesMatrix) <- "ClustyClust"
    rownames(clusterTypesMatrix) <- paste("Cluster", rownames(clusterTypesMatrix), sep=" ")
    cellCalls = data.frame(matrix(nrow=length(rownames(clusterTypesMatrix)), ncol=1))
    colnames(cellCalls) = "CellCall"
    write.csv(rownames(clusterTypesMatrix), file = paste0(outfileDirectory, "/CellType_Dict.csv"), quote = FALSE, col.names = F, row.names = F)
  }
}

##%%%%%%%%%%%%%%%%%%%%%%%##
##%%%%%%%%%%%%%%%%%%%%%%%##

#Clean up
sessionInfo()
writeLines(capture.output(sessionInfo()), paste0(outfileDirectory, "/RsessionInfo.txt"))
writeLines(" Finished: Script. Returning results back to SeqGeq.")

# save(import, file = paste0(outfileDirectory,"SeqGeq_Seurat_Object.rda"), compress = TRUE)

# } else {
#
#   writeLines("Data type not recognized.")
#
# }